#!/usr/bin/perl

use strict;

# I can't be arsed commenting this properly. View it as a challenge.

# Note to self: \e[x;xr sets the scrolling region
#               \e7 stores cursor position
#               \e8 restores cursor position
#               \e[x;xH sets cursor position

use IO::Select;
use IO::Socket;
use Sys::Hostname;
use Socket;
use FileHandle;
use IPC::Open2;
use IPC::Open3;
use Text::Wrap;
use User::pwent;

#use POSIX qw(:sys_wait_h :signal_h);
use POSIX;

my $home = $ENV{HOME};

$home = "/home/music" unless defined $home;

# Configuration options
my $maxqueuesize = 50;
my $netmask = [ 255, 255, 255, 0 ];

my $mpg123  = ["/usr/bin/mpg123", "-o", "oss", "--rate", "44100", "--stereo", "-b", "1204"];
my $mpg321 = ["/usr/local/bin/mpg321", "-o", "oss", "--rate", "44100", "--stereo", "-b", "1024"];
my $tracker = ["/usr/bin/tracker"];
my $play    = ["/usr/bin/play"];
my $ogg123  = ["/usr/bin/ogg123","-b", "10","-d","oss"];
my $ogg_stream = ["/usr/local/bin/ogg-stream"];

my $musicindex = $home . "/usr/bin/musicindex";

my $identport = 113;
#my $identport = 7779;

my $global = 0;
my $serverroot = "$home/usr/lib/puke";
my $usersfile = "$serverroot/serverusers.dat";
my $index   = "$serverroot/directories";
my $banned = "$serverroot/banned";
my $volume  = sub {
   system("echo -e \"vol Vol $_[0]\nvol Pcm 95\" | $home/usr/bin/smixer -s -");
#  system("/home/dan/usr/bin/cmdmixer",  $_[0]);
};

my $maxvolume = 95;
my $maxnamelen = 30;

my %players = ( "^.*\\.mp3\$"  => $mpg123,
	        "^.*\\.mod\$"  => $tracker,
		"^.*\\.au\$"   => $play,
		"^.*\\.wav\$"  => $play,
		"^.*\\.aiff\$" => $play,
		"^.*\\.hcom\$" => $play,
		"^.*\\.sf\$"   => $play,
		"^.*\\.voc\$"  => $play,
		"^.*\\.ogg\$"  => $ogg123,
		"^.*\\.smp\$"  => $play,
		"^.*\\.maud\$" => $play,
		"^.*\\.vwe\$"  => $play,
		"^.*\\.ogs\$" => $ogg_stream
	      );

# Commands, etc
my %command = ( say   => { 
			  subr   => \&comm_say,
			  syntax => ".say <message>",
			  help   => "Says the given message to all users" 
			 },
		tell  => {
			  subr   => \&comm_tell,
			  syntax => ".tell <who> <what>",
			  help   => "Sends the <what> privately to <who>"
			 },
		emote => {
			  subr   => \&comm_emote,
			  syntax => ".emote <action>",
			  help   => "Emotes <action> to all users"
			 },
		quit  => {
			  subr   => \&comm_quit,
			  syntax => ".quit",
			  help   => "Exits the music server"
			 },
		who   => {
			  subr   => \&comm_who,
			  syntax => ".who",
			  help   => "Tells you who's logged on"
			 },
		rename => {
			   subr   => \&comm_rename,
			   syntax => ".rename <name>",
			   help   => "Changes your name to <name>"
			  },
		help   => {
			   subr   => \&comm_help,
			   syntax => ".help",
			   help   => "Give that person a coconut!"
			  },
		colour => {
			   subr   => \&comm_colour,
			   syntax => ".colour <on|off>",
			   help   => "Turns the coloured display on or off"
			  },
		uptime => {
			   subr   => \&comm_uptime,
			   syntax => ".uptime",
			   help   => "Reports music server uptime"
			  },
		pretty => {
			   subr   => \&comm_pretty,
			   syntax => ".pretty <on|off>",
			   help   => "Turns the prettier display on or off",
			  },

		play   => {
			   subr   => \&comm_play,
			   syntax => ".play <regexp>",
			   help   => "Queues all tracks matching the regular expression"
			  },
		list   => {
			   subr   => \&comm_list,
			   syntax => ".list [regexp]",
			   help   => "Lists all tracks matching the regular expression"
			  },
		delete => {
			   subr   => \&comm_delete,
			   syntax => ".delete <regexp>",
			   help   => "Deletes all tracks belonging to you that match the regexp"
			  },
		queue  => {
			   subr   => \&comm_queue,
			   syntax => ".queue",
			   help   => "Lists the current track queue"
			  },
		random => {
                           subr   => \&comm_random,
                           syntax => ".random",
                           help   => "Randomise the current track queue"
                          },
		stop   => {
			   subr   => \&comm_stop,
			   syntax => ".stop",
			   help   => "Stops the track that's playing"
			  },
		vol    => {
			   subr   => \&comm_volume,
			   syntax => ".vol [percent]",
			   help   => "Sets or reports the volume level"
			  },
		maxvol => {
		            subr   => \&comm_maxvol,
			    syntax => ".maxvol [percent]",
			    help   => "Sets the maximum allowed volume level",
			    suid   => 1
			   },

		reload => {
			   subr   => \&comm_reload,
			   syntax => ".reload",
			   help   => "Reloads the track database",
			  },
		reinit => {
			   subr   => \&comm_reinit,
			   syntax => ".reinit",
			   help   => "Reinitalises the server",
			   suid   => 1
			  },
		restart => {
			    subr   => \&comm_restart,
			    syntax => ".restart",
			    help   => "Kills the server with an abnormal exit code (will force a server restart)",
			    suid   => 1
			   },
		shutdown => {
			     subr   => \&comm_shutdown,
			     syntax => ".shutdown",
			     help   => "Shuts down the server",
			     suid   => 2
			    },
		add    => {
			   subr   => \&comm_add,
			   syntax => ".add",
			   help   => "Checks your ~/music directory for new tracks",
			  },

		kick  => {
			  subr => \&comm_kick,
			  syntax => ".kick <who> [message]",
			  help   => "Kicks <who> off the server",
			  suid   => 1
			 },
		promote => {
			    subr => \&comm_promote,
			    syntax => ".promote <who>",
			    help   => "Promotes <who> to a superuser",
			    suid   => 1
			   },
	        demote  => {
			    subr => \&comm_demote,
			    syntax => ".demote <who>",
			    help   => "Demotes <who> to a normal user",
			    suid   => 1
			   },
		split   => {
			    subr => \&comm_split,
			    syntax => ".split <who> <whoelse>",
			    help   => "Turns <who> into two users - <who> and <whoelse>",
			    suid   => 1
			   },
		urename => {
			    subr => \&comm_urename,
			    syntax => ".urename <who> <whoelse>",
			    help   => "Renames <who> to <whoelse>",
			    suid   => 1
			   },
		murder  => {
			    subr => \&comm_murder,
			    syntax => ".murder <who>",
			    help   => "Deletes <who> from the server",
			    suid   => 1
			   },
		kickban => { 
			    subr => \&comm_kickban,
			    syntax => ".kickban <who> [message]",
			    help   => "Kicks <who> off and bans them",
			    suid   => 1
			   },
		ban     => {
			    subr => \&comm_ban,
			    syntax => ".ban <who>",
			    help   => "Bans <who> from the server",
			    suid   => 1
			   },
		unban   => {
			    subr => \&comm_unban,
			    syntax => ".unban <who>",
			    help   => "Allows <who> to log on to the server again",
			    suid   => 1
			   },
		listacc => {
			    subr => \&comm_listacc,
			    syntax => ".listacc",
			    help   => "Produces a list of all known users",
			    suid   => 1
			   },
		noplay => {
		   subr => \&comm_noplay,
		   syntax => ".noplay <regex>",
		   help => "Bans all tracks matching regex",
		   suid => 1,
		},
		pause => {
			    subr => \&comm_pause,
			    syntax => ".pause",
			    help   => "Temporarily pauses playback"
		},
		resume => {
			    subr => \&comm_resume,
			    syntax => ".resume ",
			    help   => "Resumes previously paused playback"
		}
		);

my @intromsg;
my @helpmsg;
my @suidhelp;

# Current status

my $time_up = 0;

my %sockets;    # Maps sockets -> (socket object, data, subroutine)
my %connection; # Maps sockets -> (socket object, host name, user name)
my %users;      # Maps user names -> (aliases, superuser status, status flag, connection no, connection object)
my %superusers = ( dan_1 => 2, dan_2 => 2, dan_3 => 2, matthew => 2, dturner => 2 );
my %aliases = ( esvce => "superuser",
		ahunter => "(zombie)" );
my %usecolour;   # User IDs to use ANSI colour
my %usepretty;   # User IDs to use extra pretty display
my %nousers;
my %userids = ( "superuser" => "esvce",
	        "(zombie)" => "ahunter" );
my %lasttime;
my %lastcount;

my %colours = ( tell     => "\e[0;1;36m",
		telldest => "\e[0;36m",
		say      => "\e[0m",
		sysmsg   => "\e[0;1;32m",
		help     => "\e[0;33m",
		queue    => "\e[0;32m",
		response => "\e[0;1;31m",
		kick     => "\e[0;1;31m");

my %tracks;  # filename => real track name
my @dirlist; # list of directories that we search for tracks in
my %banned;  # filename => 1

#my %tracks = (
#	      "/tmp/hmm/foo.mp3" => "Actually_works/foo.mp3",
#	      thingie => "A_track_thing/track",
#	      otherthingie => "Another_track_thing/more_track",
#	      morethingie => "Whoa/Wow",
#	      bouncey => "Bouncy_People/Happy_Bouncy_Bouncing_on_Bouncy_Castles"
#	     );

my @queue;         # Current queue of tracks
my @queuedby;      # The uid of the queuer
my $playing = "";  # The track currently playing
my $playedby = ""; # Who played it

my $vol = 40;      # Volume in %

my ($childpid, $childout, $childin, $paused);

my $listenport = 7778;
my $listener;
my $ourip;

sub logit
  {
    print STDERR "$_[0]\n";
  }

sub ping
  {
    # Build up select structures

    my($bits, $bobs);
    
    $bits = new IO::Select();
    $bobs = new IO::Select();

    for (values(%sockets))
      {
	my($socket) = @$_[0];
	$bits->add($socket);
	$bobs->add($socket);
      }

    # Do the select
    my @results = IO::Select::select($bits, undef, $bobs);
    my ($read, $erm) = ($results[0], $results[2]);

    # Report things
    for (@$read)
      {
	my $no = $_->fileno();
	my $sock = $sockets{$no};
	my ($obj, $data, $subr) = @$sock;

	if ($subr)
	  {
	    &$subr($data, $_, "READ");
	  }
	else
	  {
	    delete $sockets{$no};
	  }
      }

    for (@$erm)
      {
	my $no = $_->fileno();
	my $sock = $sockets{$no};
	my ($obj, $data, $subr) = @$sock;

	&$subr($data, $_, "OUCH");
      }
  }

my %io_buffer;
my %io_pending;

sub pip
  {
    # We have a connection... What to do?
    my ($data, $obj, $action) = @_;

    if ($action eq "OUCH")
      {
	die "Lost listening socket";
      }

    logit "Server: Accepting client connection";
    my $newsock = $listener->accept();
    my $connfrom = $newsock->peerhost();
    logit "Server: Connection from $connfrom";

    my $fn = $newsock->fileno;
    delete $io_buffer{$fn};
    delete $io_pending{$fn};

if (0) {
    if ($connfrom =~ /^([0-9]+)\.([0-9]+)\.([0-9]+)\.([0-9]+)$/)
      {
	my $maskedconn = [ $1 & @$netmask[0], $2 & @$netmask[1], $3 & @$netmask[2], $4 & @$netmask[3] ];

	for (my $x = 0; $x<4; $x++)
	  {
	    if ((@$ourip[$x] & @$netmask[$x]) != @$maskedconn[$x] &&
	        $connfrom != "127.0.0.1")
	      {
		logit "Server: Connection to client refused (not local, so identd won't be sensible)";

		close $newsock;
		return;
	      }
	  }
      }
    else
      {
	close $newsock;
	return;
      }
}
    print $newsock "|===--- Slightly Smelly Server ---===|\n";

    my $no = $newsock->fileno();
    $connection{$no} = [$newsock, $newsock->peerhost(), undef];

    print $newsock "Finding out who you are.";
    flush $newsock;

    my $identsock = new IO::Socket::INET(PeerAddr => $newsock->peerhost(),
					 PeerPort => $identport,
					 Proto    => 'tcp',
					 Timeout  => 5);

    unless($identsock)
      {
	print $newsock ".. anonymous. Go away\n";
	logit "Server: Cannot connect to identd on calling machine";
	delete $connection{$no};
	return;
      }

    logit "Server: Connected to identd";

    my $port = $newsock->peerport();
    print $identsock "$port,$listenport\n";

    print $newsock ".";
    flush $newsock;

    my $identno = $identsock->fileno();
    my @identdata = ($newsock, $no, $identno);
    $sockets{$identno} = [$identsock, \@identdata, \&gotid];
  }

sub gotid
  {
    my ($data, $obj, $action) = @_;
    my ($newsock, $no, $identno) = @$data;
print "Ident: $data $obj $action\n";
    if ($action ne "OUCH" && !$obj->eof())
      {
	while (<$obj>)
	  {
	    if (/[0-9]+\s*,\s*[0-9]+\s*:\s*USERID\s*:.*:\s*([A-z0-9]+)\s*$/)
	      {
		# Amend our information about newsock & connect whoever it is
		my $conn = $connection{$no};
		my ($so, $hostname, $un) = @$conn;
		
		my $username = ($global ? "$1\@$hostname" : $1);

		$connection{$no} = [$newsock, $hostname, $username];
		$sockets{$no} = [$newsock, [$newsock, $no, $username], \&burp];

		logit "Client: $username has connected";
		print $newsock ". Hello $1 at $hostname\n";
		flush $newsock;

		if ($users{$username} ne "")
		  {
		    my $user = $users{$username};
		    my ($alias, $superuser, $stat, $num) = @$user;

		    print $newsock "You already seem to be connected ";

		    if ($stat != 0)
		      {
			print $newsock "so I'm booting you off\n";

			delete $connection{$no};
			delete $sockets{$no};
			delete $connection{$identno};
			delete $sockets{$identno};
			return;

			print "Hmm\n";
		      }
		    else
		      {
			print $newsock "so I'm booting the other guy off\n";

			my $conn = $connection{$num};
			my ($otherguy) = @$conn;

			delete $connection{$num};
			delete $sockets{$num};
			delete $users{$1};
		      }
		  }

		my $alias = $aliases{$username};

		$users{$username} = [$alias, $superusers{$username}, 0, $no, $newsock];

		if ($alias eq "")
		  {
		    print $newsock "I haven't heard from you before.\nWhat do you wish to be known as? ";
		    flush $newsock;

		    $users{$username} = [$alias, $superusers{$username}, 1, $no, $newsock];
		  }
		else
		  {
		    if ($nousers{$username} > 1)
		      {
			print $newsock "You have more than one account registered to you\nWhich one do you want to use? ";
			flush $newsock;

			$users{$username} = [$alias, $superusers{$username}, 2, $no, $newsock];
		      }
		    else
		      {
			logit "Client: $username is known as $alias";
			
			my $connectmsg = "--- $alias has connected";

			intro($username, $newsock);

			if ($superusers{$username} < 0)
			  {
			    logit "Client: $username is banned";

			    sayto($username, "--- You are banned. Go away", 'sysmsg');
			    delete $users{$username};
			    delete $sockets{$no};
			    delete $connection{$no};

			    $connectmsg = "";
			  }
			elsif ($superusers{$username} == 1)
			  {
			    logit "Client: $username has super user status " . $superusers{$username};
			    $connectmsg = "*** $alias (superuser) has connected";
			  }
			elsif ($superusers{$username} == 2)
			  {
			    logit "Client: $username has ultra user status " . $superusers{$username};
			    $connectmsg = "*** $alias (uberservercommandant) has connected";
			  }
			if ($connectmsg ne "")
			  {
			    saynot($username,$connectmsg,'sysmsg');
			  }
		      }
		  }

		    # Get rid of the identd socket (it'll error if we get to EOF)
		    delete $connection{$identno};
		    delete $sockets{$identno};
	      }
	    elsif (/[0-9]+\s*,\s*[0-9]+\s*:\s*ERROR\s*:\s*(.*)$/)
	      {
		print $newsock ". Go away\n";

		logit "Client: identd error - $1";

		delete $connection{$identno};
		delete $sockets{$identno};
		delete $connection{$no};
		delete $sockets{$no};		
	      }
	      last;
	  }
      }
    else
      {
	print $newsock ". Uh? Identd died\n";

	logit "Client: identd died";

	delete $connection{$no};
	delete $sockets{$no};
	delete $connection{$identno};
	delete $sockets{$identno};
      }
  }

sub saynot
  {
    my($dontsay, $message, $colour) = @_;
    
    for(values(%connection))
      {
	my($obj, $host, $user) = @$_;
	my $no = $obj->fileno();
	my $us = $users{$user};
	my ($alias, $suid, $stat) = @$us;

	if ($user ne $dontsay && $stat == 0)
	  {
	    my $realmessage = $message;

	    if ($usecolour{$user})
	      {
		$realmessage = "$colours{$colour}$message\e[0m";
	      }

	    if ($usepretty{$user})
	      {
		$realmessage = "\e7\e[2;20r\e[20H\n$realmessage\e[22;25r\e8\e[0m";
	      }
	    else
	      {
		$realmessage = "$realmessage\n";
	      }

	    print $obj "$realmessage";
	    flush $obj;
	  }
      }
  }

sub sayto
  {
    my($sayto, $message, $colour) = @_;
    my $user = $users{$sayto};

    if ($user ne "")
      {
	my ($alias, $suid, $stat, $num, $obj) = @$user;

	if ($usecolour{$sayto} == 1)
	  {
	    $message = "$colours{$colour}$message\e[0m";
	  }

	if ($usepretty{$sayto} == 1)
	  {
	    $message = "\e7\e[2;20r\e[20H\n$message\e[22;25r\e8\e[0m";
	  }
	else
	  {
	    $message = "$message\n";
	  }

	print $obj "$message";
	flush $obj;
      }
  }

sub floodcheck
{
  my($userid) = @_;
  my $thistime = time();

  $thistime = $thistime / 10;

  if ($lasttime{$userid} == $thistime)
    {
      $lastcount{$userid}++;
      if ($lastcount{$userid} > 10)
	{
	  if ($superusers{$userid} <= 0)
	    {
	      saynot($userid, "--- $aliases{$userid} auto-kicked (Flooding)",'kick');
	      logit("Server: $aliases{$userid} auto-kicked ($userid) (Flooding)");
	      kick($userid);
	    }
	  else
	    {
	      sayto($userid, "*** You are flooding", 'kick');
	    }
	}
    }
  else
    {
      $lasttime{$userid} = $thistime;
      $lastcount{$userid} = 0;
      if ($lastcount{$userid} > 2)
	{
	  logit("Server: $aliases{$userid} may be flooding ($lastcount{$userid})");
	}
    }
}


sub receive
  {
    my $sock = shift;
  
    my $fh = $sock->fileno;
  
    my $buffer = '';
    my $rv = sysread($sock, $buffer, 1024);
    if (!defined($rv)) {
      if ($! == EAGAIN) {
        # would block
        return [];
      } else {
        return undef;
      }
    } else {
      if ($rv) {
        # successfully read $rv bytes from HANDLE
        $io_buffer{$fh} .= $buffer;
      } else {
        # End of file
        return undef;
      }
    }
    
    my @lines;
    
    # Chomp carridge returns
    $io_buffer{$fh} =~ s/\r//g;
    # XXX fix me to use a regex in array context or split
    while ((my $pos = index $io_buffer{$fh}, "\n") != -1) {
      push @lines, (substr $io_buffer{$fh}, 0, $pos) . "\r\n";
      $io_buffer{$fh} = substr $io_buffer{$fh}, $pos+1;
      $io_buffer{$fh} = '' unless defined $io_buffer{$fh};
    }
    
    push @{$io_pending{$fh}}, @lines;
  }



sub burp
  {
    my ($data, $obj, $action) = @_;
    my ($newsock, $no, $userid) = @$data;

    my $fn = $newsock->fileno;
    $io_buffer{$fn} = '' unless exists $io_buffer{$fn};
    $io_pending{$fn} = [] unless exists $io_pending{$fn};

    if ($action eq "OUCH")
      {
	saynot($userid, "--- Connection lost to $aliases{$userid}",'sysmsg');
	logit "Client: $userid connection lost";

	delete $connection{$no};
	delete $sockets{$no};
	delete $users{$userid};
      }
    else
      {
        my $status = &receive($newsock);
	
	if (! defined $status) {
  	  saynot($userid, "--- Connection lost to $aliases{$userid}",'sysmsg');
	  logit "Client: $userid connection lost";

	  delete $connection{$no};
	  delete $sockets{$no};
	  delete $users{$userid};
	}
	while ($#{$io_pending{$fn}} != -1) {
	  my $line = shift @{$io_pending{$fn}};
	 # warn "Before $line\n";
         # $line =~ s/\p{IsCntrl}//g;
	 # warn "After $line\n";
	  floodcheck($userid);

	while ($line =~ /^(.*)\s%([1-~]+)%\s*(.*)$/)
	  {
	    my $alias = $aliases{$2};
	    $line = "$1 $alias $3";
	  }

	my $user = $users{$userid};
	if ($user eq "")
	  {
	    return;
	  }
	my ($alias, $superuser, $stat, $num) = @$user;

	if ($stat == 0)
	  {
	    # Command mode
	    # Abbreviations
	    if ($line =~ /^\'(.*)$/)
	      {
		$line = ".say \"$1\"";
	      }
	    elsif ($line =~ /^:(.*)$/)
	      {
		$line = ".emote $1";
	      }
	    elsif ($line =~ /^>([^\s]+)\s+(.*)$/)
	      {
		$line = ".tell $1 $2";
	      }
	    elsif ($line =~ /^\s*([^.].*)\s+$/)
	      {
		$line = ".say \"$1\"";
	      }

	    # Commands
	    if ($line =~ /^\s*\.([A-z]+)\s+(.*)$/)
	      {
		my $lower    = $1;
		$lower =~ tr/[A-Z]/[a-z]/;
		my $comm = $command{$lower};
		my $commfunc = $$comm{'subr'};

		if ($commfunc eq "" || $superuser < $$comm{'suid'} )
		  {
		    sayto($userid, "*** Unknown command", 'response');
		    flush $obj;
		  }
		else
		  {
		    eval {
		      &$commfunc($data, $2);
		    };
		    if ($@)
		      {
			print $obj "!!! Broken command\n";
			logit "Error: $lower - command broken";
			logit "Ooops: $@";
		      }
		  }
	      }
	  }
	elsif ($stat == 1)
	  {
	    # Alias prompt mode
	    if ($line =~ /^\s*([1-~]*)\s*$/)
	      {
		if (length($1) <= $maxnamelen && length($1) > 1)
		  {
		    if ($userids{lc $1} eq "")
		      {
			$aliases{$userid} = $1;
			$users{$userid} = [$1, $superusers{$userid}, 0, $no, $obj];
			$userids{lc $1} = $userid;
			
			logit "Client: $userid is now known as $1";
			
			save_users();
			
			saynot($userid, "--- $1 has connected",'sysmsg');
			saynot($userid, "--- $1 is new, so be nice to him",'sysmsg');
			intro($userid, $newsock);
		      }
		    else
		      {
			print $obj "That name has already been used. Please re-enter: ";
			flush $obj;
		      }
		  }
		else
		  {
		    print $obj "That name is too long. Please re-enter: ";
		    flush $obj;
		  }
	      }
	    else
	      {
		print $obj "That is not a valid name. Please re-enter: ";
		flush $obj;
	      }
	  }
	elsif ($stat == 2)
	  {
	    if ($line =~ /^\s*([ -~]+)\s*$/)
	      {
		my $realid = $userids{lc $1};
		my $newalias = $1;

		logit "Client: User $userid now known as $1 ($realid)";
		
		my $match;
		if ($global) {
		  $userid =~ /^(.*)\@(.*)$/;
		  my $n = $1;
		  my $h = $2;
		  $match = "$n\_[0-9]\@$h";
		} else {
		  $match = "$userid\_[0-9]";
		}
		
		if ($realid !~ /^$match$/)
		  {
		    print $obj "That userID does not belong to you\n";
		    print $obj "Please re-enter: ";
		    flush $obj;
		  }
		else
		  {
		    if ($users{$userids{lc $newalias}} ne "")
		      {
			print $obj "You already seem to be connected, so I'm booting the other guy off\n";

			my $user = $users{$userids{lc $newalias}};
			my ($alias, $superuser, $stat, $num, $obj) = @$user;

			my $conn = $connection{$num};
			my ($otherguy) = @$conn;

			print $obj "You've been displaced\n";

			delete $connection{$num};
			delete $sockets{$num};
			delete $users{$userids{lc $newalias}};
		      }
		    
		    delete $users{$userid};

		    my $conn = $connection{$no};
		    my ($x, $hostname, $y) = @$conn;

		    $users{$realid} = [$newalias, $superusers{$realid}, 0, $no, $newsock];
		    $connection{$no}  = [$obj, $hostname, $realid];
		    $sockets{$no}     = [$obj, [$obj, $no, $realid], \&burp];

		    my $signmess = "$newalias has connected";
		    
		    if ($superusers{$realid} == -1)
		      {
			$signmess = "";
			sayto($realid, "--- You are banned. Go away", 'sysmsg');
			logit "Client: $newalias is banned";

			delete $users{$realid};
			delete $connection{$no};
			delete $sockets{$no};
		      }
		    elsif ($superusers{$realid} == 1)
		      {
			$signmess = "$newalias (superuser) has connected";
		      }
		    elsif ($superusers{$realid} == 2)
		      {
			$signmess = "$newalias (uberservercommandant) has connected";
		      }
		    if ($signmess ne "")
		      {
			saynot($realid, "--- $signmess",'sysmsg');
			intro($realid, $newsock);
		      }
		  }
	      }
	  }
	  }
      }
  }


sub comm_noplay
{
   my($data, $param) = @_;
   my($newsock, $no, $userid) = @$data;
   my $user = $users{$userid};

    my $regexp = "^.*\$";

    if ($param =~ /^\s*([^\s]+[ -~]*)\s*$/)
      {
	$regexp = $1;
      }
    else
      {
	sayto($userid, "*** Not a track name", 'response');
	return;
      }

    my $matched = 0;

    for(sort {$a cmp $b } keys(%tracks))
      {
	my $name = $tracks{$_};
	
	if ($name =~ /$regexp/i) {	
	    if (exists $banned{$_}) {
	      logit("Error: $_ is already banned");
	      sayto($userid, "*** $name is already banned", 'response');
	    } else {
 	      saynot("", "*** $userid has banned track $name", 'sysmsg');
	      $banned{$_} = 1;
	      $matched++;
	    }
	}
      }

    if ($matched <= 0)
      {
	sayto($userid, "*** No tracks match that pattern", 'response');
      }
  save_banned();
}


sub comm_say
  {
    my($data, $param) = @_;
    my($newsock, $no, $userid) = @$data;

    my $spaces = "";
    my $who = $aliases{$userid};

    $Text::Wrap::columns = 74;

    if ($param =~ /^\s*\"(.*)\"\s*$/)
      {
	$param = $1;
      }

    if ($param =~ /^\s*$/)
      {
	return;
      }

    for (my $i = 0; $i<((1+$maxnamelen)-length($who)); $i++)
      {
	$spaces .= " ";
      }

    my $message = "$who$spaces:$param";
    saynot("", 
	   wrap("$who$spaces:", 
		"             ",
		$param),
	   'say');
  }

sub comm_emote
  {
    my($data, $param) = @_;
    my($newsock, $no, $userid) = @$data;

    my $message = "$aliases{$userid} $param";
    saynot("", $message,'emote');
  }

sub comm_quit
  {
    my($data, $param) = @_;
    my($newsock, $no, $userid) = @$data;

    saynot($userid, "--- $aliases{$userid} has quit",'sysmsg');
    sayto($userid, "*** Toodle-oo",'sysmsg');

    logit("Client: $userid has quit");
    
    if ($usepretty{$userid}) {
      unset_pretty($newsock);
    }
    kick($userid);
  }

sub kick
  {
    my($userid) = @_;

    my $user = $users{$userid};

    if ($user ne "")
      {
	my ($alias, $suid, $stat, $conn, $obj) = @$user;

	delete $connection{$conn};
	delete $sockets{$conn};
	delete $users{$userid};

	if ($usepretty{$userid})
	  {
	    unset_pretty($obj);
	  }

	$obj->close();
	flush $obj;
      }
    else
      {
	logit "Error: Tried to kick() an unknown user";
      }
  }

sub comm_help
  {
    my($data, $param) = @_;
    my($newsock, $no, $userid) = @$data;
    my $user = $users{$userid};
    my($alias, $suid) = @$user;

    for (@helpmsg)
      {
	sayto($userid, "$_", 'help');
      }

    if ($suid > 0)
      {
	for (@suidhelp)
	  {
	    sayto($userid, "$_", 'help');
	  }
      }

    sayto($userid, "*** Available commands", 'help');
    sayto($userid, "***", 'help');

    for (sort(keys(%command)))
      {
	my $com = $command{$_};
	if ($suid >= $$com{'suid'} && $$com{'syntax'} ne "")
	  {
	    my $spaces = "";
	    for (my $i = 0; $i < (18 - length($$com{'syntax'})); $i++)
	      {
	        $spaces .= " ";
	      }

	    sayto($userid, "*** $$com{'syntax'}$spaces $$com{'help'}", 'help');
	  }
      }
  }

sub comm_colour
  {
    my($data, $param) = @_;
    my($newsock, $no, $userid) = @$data;

    if ($param =~ /on/i)
      {
	$usecolour{$userid} = 1;
	sayto($userid, "*** Colour switched on", 'response');
      }
    else
      {
	$usecolour{$userid} = 0;
	sayto($userid, "*** Colour switched off", 'response');
      }

    save_users();
  }

sub comm_pretty
  {
    my($data, $param) = @_;
    my($newsock, $no, $userid) = @$data;

    if ($param =~ /on/i)
      {
	$usepretty{$userid} = 1;
	$usecolour{$userid} = 1;
	set_pretty($newsock);
	sayto($userid, "*** Pretty switched on", 'response');
      }
    else
      {
	if ($usepretty{$userid})
	  {
	    unset_pretty($newsock);
	    $usepretty{$userid} = 0;
	  }
	sayto($userid, "*** Pretty switched off", 'response');
      }

    save_users();
  }

sub comm_tell
  {
    my($data, $param) = @_;
    my($newsock, $no, $userid) = @$data;

    if ($param =~ /^\s*([^\s]+)\s+(.*)$/)
      {
	my $uid = $userids{lc $1};
	my $user = $users{$uid};

	if ($user ne "")
	  {
	    my $spaces = "";
	    my $spaces2 = "";
	    my $who = $aliases{$userid};
	    my $who2 = $aliases{$uid};
	    
	    for (my $i = 0; $i<((1+$maxnamelen)-length($who)); $i++)
	      {
		$spaces .= " ";
	      }
	    	    
	    for (my $i = 0; $i<((1+$maxnamelen)-length($who2)); $i++)
	      {
		$spaces2 .= " ";
	      }

	    $Text::Wrap::columns = 74;

	    my $message = wrap("$who$spaces>",
			       "             ",
			       "$2");
	    my $message2 = wrap("$who2$spaces2<",
			       "             ",
			       "$2");

	    sayto($userid, "$message2", 'telldest');
	    sayto($uid, "$message", 'tell');
	  }
	else
	  {
	    sayto($userid, "*** $1 is not logged on", 'response');
	  }
      }
    else
      {
        sayto($userid, "*** Syntax error", 'response');
      }
  }

sub comm_split
{
    my($data, $param) = @_;
    my($newsock, $no, $userid) = @$data;
    my $user = $users{$userid};

    if ($user ne "")
      {
	my ($alias, $suid, $stat, $conn, $obj) = @$user;

	if ($suid <= 0)
	  {
	    sayto($userid, "*** Only superusers can split people", 'response');
	    return;
	  }
      }
    else
      {
	return;
      }

    if ($param =~ /^\s*([1-~]+)\s*([1-~]+)\s*$/)
      {
	my $who = $1;
	my $whoto = $2;
	my $uid = $userids{lc $who};

	$who = $aliases{$uid};

	if ($userids{lc $whoto} ne "")
	  {
	    sayto($userid, "*** Destination user already exists", 'response');
	    return;
	  }
	if ($uid eq "")
	  {
	    sayto($userid, "*** Original user does not exists", 'response');
	    return;
	  }

	my $topuid = $uid;

	if ($uid =~ /^(.*)\_[0-9]/)
	  {
	    $topuid = $1;
	  }

	logit("Server: Splitting $topuid into $whoto and $who");
	if ($uid eq $topuid)
	  {
	    sayto($uid, "*** You're being split. Byeee",'kick');
	    saynot($uid, "*** $who is hit by an enormous axe and splits in two",'kick');
	    kick $uid;

	    $nousers{$uid} = 2;
	    $aliases{"$uid\_1"} = $aliases{$uid};
	    $userids{lc $who} = "$uid\_1";
	    $aliases{"$uid\_2"} = $whoto;
	    $userids{lc $whoto} = "$uid\_2";
	    $aliases{$uid} = $uid;

	    $superusers{"$uid\_1"} = $superusers{$uid};
	    $superusers{"$uid\_2"} = 0;

	    save_users();
	  }
	else
	  {
	    my $num = $nousers{$topuid} + 1;
	    
	    if ($num>=10)
	      {
		sayto($userid, "*** Any one user can only have 9 other names", 'response');
		return;
	      }

	    $nousers{$topuid} = $num;
	    $aliases{"$topuid\_$num"} = $whoto;
	    $userids{lc $whoto} = "$topuid\_$num";
	    $superusers{$userids{lc $whoto}} = $superusers{$uid};

	    sayto($userid, "*** $topuid can now log on as $whoto", 'response');

	    save_users();
	  }
      }
    else
      {
	sayto($userid, "*** Syntax: split <username> <new user>", 'response');
      }
}

sub comm_promote
{
    my($data, $param) = @_;
    my($newsock, $no, $userid) = @$data;
    my $user = $users{$userid};

    if ($user ne "")
      {
	my ($alias, $suid, $stat, $conn, $obj) = @$user;

	if ($suid <= 0)
	  {
	    sayto($userid, "*** Only superusers can promote people", 'response');
	  }
	else
	  {
	    if ($param =~ /^\s*([^\s]+)\s*$/)
	      {
		my $user = $users{$userids{lc $1}};
		my $uid  = $userids{lc $1};
		my $who  = $aliases{$uid};

		if ($uid ne "" && $superusers{$uid} == 0)
		  {
		    $superusers{$userids{lc $who}} = 1;
		    sayto($userid, "*** $who promoted", 'response');
		    logit "Server: $userids{lc $who} promoted by $aliases{$userid}";
		    save_users();
		  }
		else
		  {
		    sayto($userid, "*** Can't promote $who (already superuser or unknown user", 'response');
		  }
		if ($user ne "")
		  {
		    my ($alias, $suid2, $stat, $conn, $obj) = @$user;
		   
		    if ($suid2 == 0)
		      {
			sayto($uid, "*** Congratulations on your promotion to superuser status", 'sysmsg');
			saynot($uid, "--- $who promoted to superuser status",'sysmsg', 'sysmsg');

			$users{$uid} = [$alias, 1, $stat, $conn, $obj];
		      }
		  }
	      }
	  }
      }
}

sub comm_demote
{
    my($data, $param) = @_;
    my($newsock, $no, $userid) = @$data;
    my $user = $users{$userid};

    if ($user ne "")
      {
	my ($alias, $suid, $stat, $conn, $obj) = @$user;

	if ($suid <= 0)
	  {
	    sayto($userid, "*** Only superusers can demote people", 'response');
	  }
	else
	  {
	    if ($param =~ /^\s*([^\s]+)\s*$/)
	      {
		my $user = $users{$userids{lc $1}};
		my $uid  = $userids{lc $1};
		my $who  = $aliases{$uid};

		if ($uid ne "" && $superusers{$uid} == 2)
		  {
		    sayto($userid, "*** Gerroff", 'response');
		    return;
		  }
		if ($uid ne "" && $superusers{$uid} == 1)
		  {
		    $superusers{$userids{lc $who}} = 0;
		    sayto($userid, "*** $who demoted", 'response');
		    logit "Server: $userids{lc $who} demoted by $aliases{$userid} ($userid)";
		    save_users();
		  }
		else
		  {
		    sayto($userid, "*** Can't demote $who (not superuser, or unknown user)", 'response');
		  }
		if ($user ne "")
		  {
		    my ($alias, $suid2, $stat, $conn, $obj) = @$user;
		   
		    if ($suid2 == 1)
		      {
			sayto($uid, "*** You have been demoted. Hahaha", 'sysmsg');
			saynot($uid, "--- $who demoted to pathetic status",'sysmsg');

			$users{$uid} = [$alias, 0, $stat, $conn, $obj];
		      }
		  }
	      }
	  }
      }
}

sub comm_kick
  {
    my($data, $param) = @_;
    my($newsock, $no, $userid) = @$data;
    my $user = $users{$userid};

    if ($user ne "")
      {
	my ($alias, $suid, $stat, $conn, $obj) = @$user;

	if ($suid <= 0)
	  {
	    sayto($userid, "*** Only superusers can kick people off", 'response');
	  }
	else
	  {
	    if ($param =~ /^\s*([^\s]+)\s*(.*)$/)
	      {
		my $user = $users{$userids{lc $1}};
		my $who  = $1;
		my $rider = $2;
		$rider =~ s/\s/ /g;

		if ($rider =~ /^\s*((([^\s]+)\s+([^\s]+))+)\s*$/)
		  {
		    $rider = $1;
		  }
		elsif ($rider =~ /^\s*([^\s]+)\s*$/)
		  {
		    $rider = $1;
		  }

		if ($user ne "")
		  {
		    my ($alias, $suid2, $stat, $conn, $obj) = @$user;
		    
		    if ($suid2 == 0)
		      {
			sayto($userids{lc $who}, "*** WHAM! A steel-toed boot slams into your behind", 'kick');
			if ($rider ne "")
			  {
			    sayto($userids{lc $who}, "($rider\)", 'kick');
			  }
			kick($userids{lc $who});
			logit("SuperUser: $who kicked by $aliases{$userid}");
			
			if ($rider ne "")
			  {
			    saynot($userids{lc $who}, "--- $who is kicked waaaay over the horizon ($rider\)",'kick');
			  }
			else
			  {
			    saynot($userids{lc $who}, "--- $who is kicked waaaay over the horizon",'kick');
			  }
		      }
		    else
		      {
			saynot($userids{lc $who}, "--- $aliases{$userid} gives $who a damn good kicking!",'kick');
			sayto($userids{lc $who}, "*** $aliases{$userid} gives you a damn good kicking!",'kick');
		      }
		  }
	      }
	  }
      }
  }

sub comm_ban
  {
    my($data, $param) = @_;
    my($newsock, $no, $userid) = @$data;
    my $user = $users{$userid};

    if ($user ne "")
      {
	my ($alias, $suid, $stat, $conn, $obj) = @$user;

	if ($suid <= 0)
	  {
	    sayto($userid, "*** Only superusers can ban people", 'response');
	    return;
	  }
	else
	  {
	    if ($param =~ /^\s*([1-~]+)\s*/)
	      {
		my $who = $1;
		my $uid = $userids{lc $who};

		if ($uid ne "")
		  {
		    if ($users{$uid} ne "")
		      {
			sayto($userid, "*** Kickban people who are logged on", 'response');
			return;
		      }
		    
		    if ($superusers{$uid} > 0)
		      {
			sayto($userid, "*** You can't ban superusers", 'response');
			return;
		      }
		    
		    $superusers{$uid} = -1;
		    sayto($userid, "*** $who banned", 'response');

		    logit "Client: $aliases{$userid} ($userid) banned $who";
		    save_users();
		  }
		else
		  {
		    sayto($userid, "*** That user does not exist", 'response');
		  }
	      }
	  }
      }
  }

sub comm_unban
  {
    my($data, $param) = @_;
    my($newsock, $no, $userid) = @$data;
    my $user = $users{$userid};

    if ($user ne "")
      {
	my ($alias, $suid, $stat, $conn, $obj) = @$user;

	if ($suid <= 0)
	  {
	    sayto($userid, "*** Only superusers can unban people", 'response');
	    return;
	  }
	else
	  {
	    if ($param =~ /^\s*([1-~]+)\s*/)
	      {
		my $who = $1;
		my $uid = $userids{lc $who};

		if ($uid ne "")
		  {
		    if ($users{$uid} ne "")
		      {
			sayto($userid, "*** Kickban people who are logged on", 'response');
			return;
		      }
		    
		    if ($superusers{$uid} > 0)
		      {
			sayto($userid, "*** You can't unban superusers", 'response');
			return;
		      }
		    
		    $superusers{$uid} = 0;
		    sayto($userid, "*** $who unbanned", 'response');

		    logit "Client: $aliases{$userid} ($userid) unbanned $who";
		    save_users();
		  }
		else
		  {
		    sayto($userid, "*** That user does not exist", 'response');
		  }
	      }
	  }
      }
  }

sub comm_kickban
  {
    my($data, $param) = @_;
    my($newsock, $no, $userid) = @$data;
    my $user = $users{$userid};

    if ($user ne "")
      {
	my ($alias, $suid, $stat, $conn, $obj) = @$user;

	if ($suid <= 0)
	  {
	    sayto($userid, "*** Only superusers can kick people off", 'response');
	  }
	else
	  {
	    if ($param =~ /^\s*([^\s]+)\s*(.*)$/)
	      {
		my $user = $users{$userids{lc $1}};
		my $who  = $1;
		my $rider = $2;
		$rider =~ s/\s/ /g;

		if ($rider =~ /^\s*((([^\s]+)\s+([^\s]+))+)\s*$/)
		  {
		    $rider = $1;
		  }
		elsif ($rider =~ /^\s*([^\s]+)\s*$/)
		  {
		    $rider = $1;
		  }

		if ($user ne "")
		  {
		    my ($alias, $suid2, $stat, $conn, $obj) = @$user;
		    
		    if ($suid2 == 0)
		      {
			sayto($userids{lc $who}, "*** WHAM! A steel-toed boot slams into your behind (you are banned)", 'kick');
			if ($rider ne "")
			  {
			    sayto($userids{lc $who}, "($rider\)", 'kick');
			  }
			kick($userids{lc $who});
			$superusers{$userids{lc $who}} = -1;
			logit("SuperUser: $who kickbanned by $aliases{$userid}");
			save_users();
			
			if ($rider ne "")
			  {
			    saynot($userids{lc $who}, "--- $who is kicked waaaay over the horizon ($rider\)",'kick');
			    saynot($userids{lc $who}, "--- And don't come back!",'kick');
			  }
			else
			  {
			    saynot($userids{lc $who}, "--- $who is kicked waaaay over the horizon",'kick');
			    saynot($userids{lc $who}, "--- And don't come back!",'kick');
			  }
		      }
		    else
		      {
			saynot($userids{lc $who}, "--- $aliases{$userid} gives $who a damn good kicking!",'kick');
			sayto($userids{lc $who}, "*** $aliases{$userid} gives you a damn good kicking!",'kick');
		      }
		  }
	      }
	  }
      }
  }

sub comm_rename
{
  my($data, $param) = @_;
  my($newsock, $no, $userid) = @$data;
  my $user = $users{$userid};

  if ($param =~ /^\s*([1-~]+)\s*$/)
    {
      if (length($1) <= $maxnamelen)
	{
	  if ($userids{lc $1} eq "")
	    {
	      my ($alias, $suid, $stat, $num, $obj) = @$user;
	      
	      saynot("", "--- $aliases{$userid} has changed into $1",'sysmsg');
	      
	      logit("Client: $aliases{$userid} has changed into $1");
	      
	      delete $userids{lc($aliases{$userid})};
	      $aliases{$userid} = $1;
	      $userids{lc $1} = $userid;
	      
	      $users{$userid} = [ $1, $suid, $stat, $num, $obj ];
	      
	      save_users();
	    }
	  else
	    {
	      sayto($userid, "*** That name is already in use", 'sysmsg');
	    }
	}
      else
	{
	  sayto($userid, "*** That name is too long", 'sysmsg');
	}
    }
  else
    {
      sayto($userid, "*** Invalid name", 'sysmsg');
    }
}

sub comm_urename
{
  my($data, $param) = @_;
  my($newsock, $no, $realuid) = @$data;
  my $user = $users{$realuid};

  if ($user ne "")
    {
      my ($alias, $suid, $stat, $conn, $obj) = @$user;
      
      if ($suid <= 0)
	{
	  sayto($realuid, "*** Only superusers can urename people", 'response');
	  return;
	}
    }
  else
    {
      return;
    }
  
  if ($param =~ /^\s*([1-~]+)\s*([1-~]+)\s*$/)
    {
      my $userid = $1;
      my $user = $users{$userid};

      if (length $2 > $maxnamelen) {
        sayto($realuid, "*** That name is too long", 'response');
	return;
      }

      if ($userids{lc $2} eq "" && $aliases{$userid} ne "" && $superusers{$userid} == 0)
	{
	  if ($user ne "")
	    {
	      saynot("$userid", "--- $aliases{$userid} has changed into $2",'sysmsg');
	      sayto("$userid", "*** You have been called $2", 'sysmsg');
	    }
	  else
	    {
	      sayto("$realuid", "*** $aliases{$userid} now known as $2", 'response');
	    }
	
	  logit("Client: $aliases{$userid} has been changed into $2 by $aliases{$realuid} ($realuid)");
  
	  delete $userids{lc($aliases{$userid})};
	  $aliases{$userid} = $2;
	  $userids{lc $2} = $userid;

	  if ($user ne "")
	    {
	      my ($alias, $suid, $stat, $num, $obj) = @$user;
	      $users{$userid} = [ $2, $suid, $stat, $num, $obj ];
	    }

	  save_users();
	}
      else
	{
	  sayto($realuid, "*** That name is already in use", 'response');
	}
    }
  else
    {
      sayto($realuid, "*** Invalid name", 'response');
    }
}

sub comm_uptime
{
  my($data, $param) = @_;
  my($newsock, $no, $userid) = @$data;
  my $user = $users{$userid};

  my $now = time();
  my $uptime = $now - $time_up;
  my $updays  = int($uptime/(3600*24));
  my $uphours = int($uptime/3600)%24;
  my $upmins  = int($uptime/60)%60;
  my $upsecs  = $uptime%60;

  sayto($userid, "*** Up $updays days $uphours hours $upmins minutes $upsecs seconds", 'help');
}

sub comm_who
{
  my($data, $param) = @_;
  my($newsock, $no, $userid) = @$data;
  my $user = $users{$userid};

  sayto($userid, "*** Logged on today are:", 'help');
  sayto($userid, "***", 'help');

  for (values(%users))
    {
      my($alias, $suid, $status, $conn, $obj) = @$_;
      my($thing) = lc $alias;

      if ($suid == 0)
	{
	  sayto($userid, "*** $alias ($userids{$thing})",'help');
	}
      elsif ($suid == 1)
	{
	  sayto($userid, "*** $alias ($userids{$thing}) {superuser}", 'help');
	}
      else 
        {
	  sayto($userid, "*** $alias ($userids{$thing}) {uberservercommandant}", 'help');
        }
    }
}

sub comm_listacc
{
  my($data, $param) = @_;
  my($newsock, $no, $userid) = @$data;
  my $user = $users{$userid};

  sayto($userid, "*** Available accounts are:", 'help');
  sayto($userid, "***", 'help');

  for (keys(%aliases))
    {
      my $user = $users{$_};
      my $suid = $superusers{$_};
      my $alias = $aliases{$_};
      my($thing) = lc $alias;

      if ($nousers{$_} == 0)
	{
	  if ($suid == 0)
	    {
	      sayto($userid, "*** $alias ($_)", 'help');
	    }
	  elsif ($suid > 0)
	    {
	      sayto($userid, "*** $alias ($userids{$thing}) {superuser}", 'help');
	    }
	  else
	    {
	      sayto($userid, "*** $alias ($userids{$thing}) {banned}", 'help');
	    }
	}
    }
}

sub comm_murder
  {
    my($data, $param) = @_;
    my($newsock, $no, $userid) = @$data;
    my $user = $users{$userid};

    if ($user ne "")
      {
	my ($alias, $suid, $stat, $conn, $obj) = @$user;

	if ($suid <= 0)
	  {
	    sayto($userid, "*** Only superusers can murder people", 'response');
	  }
	else
	  {
	    if ($param =~ /^\s*([^\s]+)\s*(.*)$/)
	      {
		my $user = $users{$userids{lc $1}};
		my $who  = $1;
		my $rider = $2;
		$rider =~ s/\s/ /g;

		if ($rider =~ /^\s*((([^\s]+)\s+([^\s]+))+)\s*$/)
		  {
		    $rider = $1;
		  }
		elsif ($rider =~ /^\s*([^\s]+)\s*$/)
		  {
		    $rider = $1;
		  }

		if ($user ne "")
		  {
		    my ($alias, $suid2, $stat, $conn, $obj) = @$user;
		    
		    if ($suid2 == 0)
		      {
			sayto($userids{lc $who}, "*** You have died", 'kick');
			if ($rider ne "")
			  {
			    sayto($userids{lc $who}, "($rider\)", 'kick');
			  }
			kick($userids{lc $who});
			logit("SuperUser: $who murdered by $aliases{$userid}");
			
			if ($rider ne "")
			  {
			    saynot($userids{lc $who}, "--- $who is kicked waaaay over the horizon (and splats upon landing) ($rider\)",'kick');
			  }
			else
			  {
			    saynot($userids{lc $who}, "--- $who is kicked waaaay over the horizon (and splats upon landing)",'kick');
			  }
		      }
		    else
		      {
			saynot($userids{lc $who}, "--- $aliases{$userid} gives $who a damn good kicking!",'kick');
			sayto($userids{lc $who}, "*** $aliases{$userid} gives you a damn good kicking!",'kick');
		      }
		  }

		my $uid = $userids{lc $who};
		my $alias = $aliases{$uid};

		logit "Server: Erasing $uid/$alias ($userid done it)";

		if ($superusers{$uid} > 0)
		  {
		    logit "Server: Superusers not erased";
		    sayto($userid, "*** You can't murder a superuser", 'response');
		    return;
		  }

		sayto($userid, "*** Murdering $uid/$alias", 'response');

		if ($uid =~ /^(.*)_([0-9])/)
		  {
		    my ($topuid, $unum) = ($1, $2);

		    if ($nousers{$topuid} >= 1)
		      {
			delete $aliases{$uid};
			delete $userids{lc $alias};

			my $ruid;

			# We now need to compact the usernames down
			for ($ruid=$unum; $ruid<$nousers{$topuid}; $ruid++)
			  {
			    my $foo      = $ruid + 1;
			    my $olduid = "$topuid\_$foo";
			    my $newuid = "$topuid\_$ruid";

			    $aliases{$newuid} = $aliases{$olduid};
			    $userids{lc($aliases{$newuid})} = $newuid;
			    $superusers{$newuid} = $superusers{$olduid};
			    delete $aliases{$olduid};
			    delete $superusers{$olduid};

			    my $user = $users{$olduid};
			    if ($user ne "")
			      {
				my ($alias, $superuser, $stat, $num, $obj) = @$user;
				$users{$newuid} = $user;
				delete $users{$olduid};
 
				{
				  my $conn = $connection{$num};
				  my ($newsock, $hostname, $un) = @$conn;
				  $connection{$num} = [$obj, $hostname, 
						       $newuid];
				}
				$sockets{$num} = [$obj, [$obj, $num, $newuid], 
						  \&burp];
      			      }
			  }
		      }

		    $nousers{$topuid}--;

		    if ($nousers{$topuid} == 1)
		      {
			logit("Server: all extra users gone - normal user now");
			$aliases{$topuid} = $aliases{"$topuid\_1"};
			$userids{lc $aliases{$topuid}} = $topuid;
			$superusers{$topuid} = $superusers{"$topuid\_1"};
			delete $nousers{$topuid};

			my $user = $users{"$topuid\_1"};
			if ($user ne "")
			  {
			    my ($alias, $superuser, $stat, $num, $obj) = @$user;
		    
			    $users{$topuid} = $user;
			    delete $users{"$topuid\_1"};

			    {
			      my $conn = $connection{$num};
			      my ($newsock, $hostname, $un) = @$conn;
			      $connection{$num} = [$obj, $hostname, $topuid];
			    }
			    $sockets{$num} = [$obj, [$obj, $num, $topuid], 
					      \&burp];
			  }

			delete $aliases{"$topuid\_1"};
		      }
		  }
		else
		  {
		    delete $aliases{$uid};
		    delete $userids{lc $alias};
		  }
	      }
	  }
      }

    save_users();
  }

# ======
# Music server commands

sub comm_list
  {
    my($data, $param) = @_;
    my($newsock, $no, $userid) = @$data;
    my $user = $users{$userid};

    my $regexp = "^.*\$";

    if ($param =~ /^\s*([^\s]+[ -~]*)\s*$/)
      {
	$regexp = $1;
      }


    my $matched = 0;

    for(sort values(%tracks))
      {
	my $trackname = $_;
	if ($trackname =~ /$regexp/i)
	  {
	    sayto($userid, "*** Matching tracks are:", 'help') if $matched < 1;
	    sayto($userid, "***", 'help') if $matched < 1;
	    $matched++;
	    sayto($userid, "*** $_", 'help');
	  }
      }

    if ($matched <= 0)
      {
	sayto($userid, "*** No tracks match that pattern", 'response');
      }
  }

sub comm_play
  {
    my($data, $param) = @_;
    my($newsock, $no, $userid) = @$data;
    my $user = $users{$userid};

    my $regexp = "^.*\$";

    if ($param =~ /^\s*([^\s]+[ -~]*)\s*$/)
      {
	$regexp = $1;
      }
    else
      {
	sayto($userid, "*** Not a track name", 'response');
	return;
      }

    my $matched = 0;

    for(sort {$a cmp $b } keys(%tracks))
      {
	my $name = $tracks{$_};
	
	if ($name =~ /$regexp/i) {
	    if (exists $banned{$_}) {
	      logit("Client: $user tried to queue a banned track");
	      sayto($userid, "*** Playing $name is banned", 'response');
	    } else {
  	      enqueue($userid, $_);
	      $matched++;
	    }
	  }
      }

    if ($matched <= 0)
      {
	sayto($userid, "*** No tracks match that pattern", 'response');
      }
  }

sub comm_queue
  {
    my($data, $param) = @_;
    my($newsock, $no, $userid) = @$data;
    my $user = $users{$userid};

    if ($#queue < 0)
      {
	if ($playing ne "")
	  {
	    sayto($userid, "*** Currently playing '$tracks{$playing}' (queued by $aliases{$playedby})", 'queue');
	  }
	else
	  {
	    sayto($userid, "*** The track queue is empty", 'queue');
	  }
	return;
      }

    if ($playing ne "")
      {
	sayto($userid, "*** Currently playing '$tracks{$playing}' (queued by $aliases{$playedby})", 'queue');
      }
    else
      {
	sayto($userid, "*** Currently queued tracks:", 'queue');
      }
    sayto($userid, "***", 'queue');

    my $num;
    for ($num = 0; $num <= $#queue; $num++)
      {
	my $n = $num + 1;
	sayto($userid, "*** $n\. $tracks{$queue[$num]} queued by $aliases{$queuedby[$num]}", 'queue');
      }
  }

sub comm_random
  {
    my($data, $param) = @_;
    my($newsock, $no, $userid) = @$data;
    my $user = $users{$userid};
     
    my $count = 10;
    if ($param =~ /^\s*(\d+)\s*$/) {
	$count = $1;
    }
    
    $count = 50 if $count > 50;
    
    my @tracks = values %tracks;
    
    for (my $i = 0 ; $i < $count ; $i++) {
	my $max = $#tracks;
	
	my $track = $tracks[int(rand $max)];
	
	sayto($userid, "*** $track", 'response');
    }
  }
  

sub comm_reinit
{
  my($data, $param) = @_;
  my($newsock, $no, $userid) = @$data;
  my $user = $users{$userid};

  if ($user ne "")
    {
      my ($alias, $suid, $stat, $conn, $obj) = @$user;
      
      if ($suid <= 0)
	{
	  sayto($userid, "*** Only superusers can reinitialise the server", 'response');
	  return;
	}
    }
  else
    {
      return;
    }

  saynot("", "*** $aliases{$userid} has requested a server reinit. One moment please...", 'sysmsg');

  read_banned();
  read_tracks();
  load_users();

  saynot("", "*** reinit complete", 'sysmsg');
}

sub comm_reload
{
  my($data, $param) = @_;
  my($newsock, $no, $userid) = @$data;
  my $user = $users{$userid};

  if ($user ne "")
    {
      my ($alias, $suid, $stat, $conn, $obj) = @$user;
      
    }
  else
    {
      return;
    }

  saynot("", "*** $aliases{$userid} has requested a track reload. One moment please...", 'sysmsg');
    
  read_tracks();

  saynot("", "*** reload complete", 'sysmsg');
}

sub comm_restart
{
  my($data, $param) = @_;
  my($newsock, $no, $userid) = @$data;
  my $user = $users{$userid};

  if ($user ne "")
    {
      my ($alias, $suid, $stat, $conn, $obj) = @$user;
      
      if ($suid <= 0)
	{
	  sayto($userid, "*** Only superusers can restart the server", 'response');
	  return;
	}
    }
  else
    {
      return;
    }

  logit "Server: $userid has requested that the server restart";

  saynot("", "*** $aliases{$userid} has shut down the server (should restart in 5-10 seconds)", 'sysmsg');
  
  kill 15, ( $childpid );
    
  exit(1);
}

sub comm_add
{
  my($data, $param) = @_;
  my($newsock, $no, $userid) = @$data;
  my $user = $users{$userid};

  my $realuid = $userid;

  if ($realuid =~ /^(.*)\_[0-9]/)
    {
      $realuid = $1;
    }

  my $pwent = getpw("$realuid");

  if ($pwent eq "")
    {
      sayto("$userid", "*** I can't find out who you are", "response");
      return;
    }

  my $dir = "@$pwent[7]/public_music/";

  for (my $x = 0; $x <= $#dirlist; $x++)
    {
      if ($dirlist[$x] eq $dir)
	{
	  sayto("$userid", "*** Your home directory is already in the list of directories that I scan", "response");
	  return;
	}
    }

  logit "Client: Adding directory $dir to list of directories to be scanned";

  $dirlist[$#dirlist + 1] = $dir;

  open DIRS, ">$index";
  for (my $x = 0; $x <= $#dirlist; $x++)
    {
      print DIRS "$dirlist[$x]\n";
    }
  close DIRS;

  sayto("$userid", "*** Added $dir to the list of directories that are scanned for music", "response");
  saynot("", "--- Reinitialising the server", "sysmsg");
  
  read_tracks();
}

sub comm_shutdown
{
  my($data, $param) = @_;
  my($newsock, $no, $userid) = @$data;
  my $user = $users{$userid};

  if ($user ne "")
    {
      my ($alias, $suid, $stat, $conn, $obj) = @$user;
      
      if ($suid <= 0)
	{
	  sayto($userid, "*** Only superusers can shut down the server", 'response');
	  return;
	}
    }
  else
    {
      return;
    }

  logit "Server: $userid has shut down the server";

  saynot("", "*** $aliases{$userid} has killed the server", 'sysmsg');
  
  kill 15, ( $childpid );
    
  exit(0);
}

sub comm_delete
  {
    my($data, $param) = @_;
    my($newsock, $no, $userid) = @$data;
    my $user = $users{$userid};
    my ($alias, $suid, $stat, $cnum, $cobj) = @$user;

    my $regexp = "^.*\$";

    if ($param =~ /^\s*([^\s]+[ -~]*)\s*$/)
      {
	$regexp = $1;
      }
    else
      {
	sayto($userid, "*** Not a track name", 'response');
	return;
      }
    

    my $matched = 0;
    my $i;

    for($i = 0; $i <= $#queue; $i++)
      {
	my $name = $tracks{$queue[$i]};
	
	if ($name =~ /$regexp/i) {
	    if ($suid <= 0 && $queuedby[$i] ne $userid)
	      {
		sayto($userid, "*** You cannot delete $name, as it was not queued by you", 'response');
	      }
	    else
	      {
		saynot("", "--- $aliases{$userid} has deleted track $name",'queue');
		logit "Client: $aliases{$userid} ($userid) has deleted track $name";
		dequeue($i);

		$i--;
	      }
	    $matched++;
	  }
      }

    if ($matched <= 0)
      {
	sayto($userid, "*** No tracks match that pattern", 'response');
      }
  }

sub music_pids
  {
    my $parent = shift;
    
    my @pids = ($parent);
    open PS, "ps -ax -o pid,ppid |";
    while (<PS>) {
      if (/^\s*(\d+)\s+$parent\s+/) {
        push @pids, $1;
      }
    }
    close PS;
    
    return @pids;
  }


sub comm_stop
  {
    my($data, $param) = @_;
    my($newsock, $no, $userid) = @$data;
    my $user = $users{$userid};
    my ($alias, $suid, $stat, $cnum, $cobj) = @$user;

    if ($playing eq "")
      {
	sayto($userid, "*** Nothing playing", 'response');

	return;
      }
    elsif ($playedby ne $userid && $suid <= 0)
      {
	sayto($userid, "*** You did not queue this track", 'response');

	return;
      }

    saynot("", "--- $aliases{$userid} has stopped the current track", 'queue');

    my @pids = &music_pids ($childpid);

    logit "Music: Killing playing process by force (stopped by $userid)";
    logit "Music: PIDS: " . join(",", @pids);

    kill 15, @pids;
  }

sub comm_maxvol
  {
    my($data, $param) = @_;
    my($newsock, $no, $userid) = @$data;
    my $user = $users{$userid};
    my ($alias, $suid, $stat, $cnum, $cobj) = @$user;

    $maxvolume = $param;

    if ($maxvolume > 100)
      {
	$maxvolume = 100;
      }
    elsif ($maxvolume < 0)
      {
	$maxvolume = 0;
      }

    $maxvolume = int($maxvolume);

    saynot("", "--- Maximum volume now $maxvolume\%", "sysmsg");
    logit "Server: $userid set the maximum volume to $maxvolume\%";

    if ($vol > $maxvolume)
      {
	$vol = $maxvolume;
	set_volume();

	saynot("", "--- Changed volume to $vol\%", "sysmsg");
      }
  }

sub comm_volume
  {
    my($data, $param) = @_;
    my($newsock, $no, $userid) = @$data;
    my $user = $users{$userid};
    my ($alias, $suid, $stat, $cnum, $cobj) = @$user;

    if ($param =~ /^\s*(\d\d?)\s*$/)
      {
	if ($1 < $maxvolume)
	  {
	    $vol = $1;
	  }
	else
	  {
	    $vol = $maxvolume;
	  }

	set_volume();

	saynot("", "--- $aliases{$userid} has set the volume to $vol\%", 'sysmsg');
      }
    elsif ($param =~ /^\s*(\+|-)(\d\d?)\s*$/) 
      {
	my $op = $1;
	my $inc = $2;
      
	if ($op eq "-") {
	    $vol -= $inc;
	} else {
	    $vol += $inc;
	}
	
	if ($vol > $maxvolume) {
	    $vol = $maxvolume;
	}
	
	set_volume();
	saynot("", "--- $aliases{$userid} has changed the volume by $op$inc to $vol\%", 'sysmsg');
      }	
    else
      {
	sayto($userid, "*** Current volume level is $vol\%", 'response');
      }
  }

sub comm_pause
  {
    my($data, $param) = @_;
    my($newsock, $no, $userid) = @$data;
    my $user = $users{$userid};
    my ($alias, $suid, $stat, $cnum, $cobj) = @$user;

    if ($playing eq '') {
      sayto($userid, "--- There's no track playing", 'response');
    } else {
      if ($paused == 1) {
        sayto($userid, "--- The track's already paused", 'response');
      } else {
        $paused = 1;
        my @pids = &music_pids ($childpid);
	logit ("Music: Pausing pids with " . POSIX::SIGSTOP . join(',', @pids));
	kill POSIX::SIGSTOP, @pids;
        saynot("", "*** $aliases{$userid} has paused playback", 'sysmsg');
      }
    }
  }

sub comm_resume
  {
    my($data, $param) = @_;
    my($newsock, $no, $userid) = @$data;
    my $user = $users{$userid};
    my ($alias, $suid, $stat, $cnum, $cobj) = @$user;

    if ($playing eq '') {
      sayto($userid, "--- There's no track playing", 'response');
    } else {
      if ($paused != 1) {
        sayto($userid, "--- The track's not paused", 'response');
      } else {
        $paused = 0;
        my @pids = &music_pids ($childpid);
	logit ("Music: Resuming pids with " . POSIX::SIGCONT . join(',', @pids));
	kill POSIX::SIGCONT, @pids;
        saynot("", "*** $aliases{$userid} has resumed playback", 'sysmsg');
      }
    }
  }



# ======
# Bits & bobs

sub set_volume
{
  &$volume($vol);
}

sub enqueue
  {
    my ($userid, $filename) = @_;

    if ($#queue >= $maxqueuesize)
      {
	sayto($userid, "*** Queue full", 'queue');
	return;
      }
    if ($tracks{$filename} eq "")
      {
	sayto($userid, "!!! Tune missing", 'queue');
	logit("Error: Tried to enqueue $filename, which has no entry in the tracks list");
	return;
      }

    my %count;
    if ($playing ne '') {
      $count{$playedby} = 1;
    }
    my $i;
    for ($i = 0 ; $i <= $#queue ; $i++) {
	if ($queue[$i] eq $filename)
	  {
	    sayto($userid, "*** $tracks{$filename} is already in the queue", 'queue');
	    return;
	  }
      my $user = $queuedby[$i];
      if (defined $count{$user}) {
        $count{$user}++;
      } else {
        $count{$user} = 1;
      }
    }

    my $priority = $count{$userid}+1 || 1;
    saynot("", "--- $aliases{$userid} has queued track \"$tracks{$filename}\"", 'queue');
    logit "Client: $aliases{$userid} ($userid) has queued \"$tracks{$filename}\"";
    
    my $doneit = 0;
    undef %count;
    if ($playing ne '') {
      $count{$playedby} = 1;
    }
    for ($i = 0 ; $i <= $#queue ; $i++) {
      my $user = $queuedby[$i];
      if (defined $count{$user}) {
        $count{$user}++;
      } else {
        $count{$user} = 1;
      }
      if ($count{$user} > $priority) {
        $doneit = 1;
	my @tail = splice @queue, $i;
	push @queue, $filename;
	push @queue, @tail;
	@tail = splice @queuedby, $i;
	push @queuedby, $userid;
	push @queuedby, @tail;
	last;
      }
    }

    if (!$doneit) {
      $queue[$#queue + 1] = $filename;
      $queuedby[$#queue] = $userid;
    }

    if ($playing eq '')
      {
	play();
      }
  }

sub dequeue
  {
    my ($tracknum) = @_;
    my $i;

    for ($i = $tracknum; $i < $#queue; $i++)
      {
	$queue[$i]    = $queue[$i+1];
	$queuedby[$i] = $queuedby[$i+1];
      }

    $#queue--;

    if ($#queue == -1)
      {
	saynot("", "--- The track queue is empty", 'queue');
      }
  }

sub set_pretty
  {
    my ($sock) = @_;

    print $sock "\e[H\e[2J\e1;1r\e[0;41;1;37m\e[1H\e[K  Music Server                                                                 ";
    print $sock "\e[2;21r\e[0;37m\e[2H";
    print $sock "\e[21;22r\e[21H\e[0;1m------------------------------------------------------------------------------";
    print $sock "\e[22;25r\e[22H";

    flush $sock;
  }

sub unset_pretty
  {
    my ($sock) = @_;

    print $sock "\e0;25r\e[r\e[H\e[2J";
    flush $sock;
  }

sub play
  {
    if ($playing ne "")
      {
	logit "Error: Tried to play track while other track still playing";
	return;
      }
    if ($#queue < 0)
      {
	logit "Error: Tried to play track with an empty queue";
	return;
      }
      
    $playing  = $queue[0];
    $playedby = $queuedby[0];

    my $player;

    foreach (keys(%players))
      {
	if ($playing =~ /$_/) {
	  $player = $players{$_};
	  last;
	}
      }

    if ($player)
      {
	logit "Music: Playing \"$tracks{$playing}\" (" . join(" ", @{$player}) . ")";
	sleep 1; # Sometimes the last task is still running
	saynot("", "--- Playing track \"$tracks{$playing}\"", 'queue');

	$childin = new FileHandle;
	$childout = new FileHandle;
	$childpid = open3($childin, $childout, '', @{$player}, $playing);
	
	my $no = $childin->fileno();
	$sockets{$no} = [ $childin, [ $childin, $no ], \&playback];
      }
    else
      {
	saynot("", "--- Can't find player for '$tracks{$playing}'", 'sysmsg');
	delete $tracks{$playing};
	$playing = "";

	dequeue(0);
	if ($#queue >= 0)
	  {
	    play();
	  }
	return;
      }

    dequeue(0);
  }

sub playback
  {
    my ($data, $obj, $action) = @_;
    my ($newsock,$no) = @$data;
    
    my $stop = 0;

    if ($action eq "OUCH")
      {
	$stop = 1;
	logit("Music: Playback terminated abnormally");
	delete $sockets{$no};
      }
    elsif ($obj->eof())
      {
	$stop = 1;
	logit("Music: Playback terminated normally");
	delete $sockets{$no};
      }

    if ($stop)
      {
	$playing = '';
	$playedby = '';
	$paused = 0;
	if ($#queue >= 0)
	  {
	    play();
	  }
	else
	  {
	    saynot("","--- Playing finished", 'queue');
	  }
      }
  }

sub read_tracks
{
  # Clear out the old tracks
  for (keys(%tracks))
    {
      delete $tracks{$_};
    }

  logit("Server: Reading tracks");

  # Read in the new
  open DIRS, "<$index";
  $#dirlist = -1;
  print "DIRS $index\n";

  while (<DIRS>)
    {
    print "Dir $_";
      if (/^([^\s]+)$/)
	{
	  my $dir = $1;
	  open FILE, "<$dir/index"
	    or open FILE, "cd $dir && $musicindex |"
	    or die "cannot open directory $dir: $!";
	  
	  $dirlist[$#dirlist + 1] = $dir;
	  
	  while (<FILE>)
	    {
	      next if /^\s*#/;
#	    print "Process $_\n";
	      if (/^(.+)$/)
		{
		  my $fullname = "$dir/$1";

		  $tracks{$fullname} = $1;
		}
	    }
	  close FILE;
	}
      else
	{
	  logit "Server: Invalid directory entry ($_)";
	}
    }
  close DIRS;
}

sub read_banned
{
  # Clear out the old tracks
  foreach (keys(%banned))
    {
      delete $banned{$_};
    }

  logit("Server: Reading banned tracks");

  # Read in the new
  open BANNED, "<$banned";
  print "BANNED $banned\n";

  while (<BANNED>)
    {
      chomp;
      print "Banned $_";
      $banned{$_} = 1;
    }
  close BANNED;
  
  foreach (keys %banned) {
    print "Got track $_";
  }
}


sub save_banned
{
  logit("Server: Saving banned tracks");
  
  open BANNED, ">$banned";
  print "BANNED $banned\n";
  
  foreach (keys %banned) {
    print BANNED "$_\n";
    print "Here $_\n";
  }
  close BANNED;
}


sub intro
{
  my ($uid,$sock) = @_;

  if ($usepretty{$uid})
    {
      set_pretty($sock);
    }

  for(@intromsg)
    {
      my $line = $_;
      if ($line =~ /^(.*)\%name\%(.*)/)
	{
	  $line = "$1$aliases{$uid}$2";
	}
      sayto($uid, "$line", 'help');
    }
}

sub save_users
{
  my $userfile;

  logit "Server: Saving users";

  open(userfile, ">$usersfile");

  for (keys(%aliases))
    {
      print userfile "UN=$_ AL=$aliases{$_} SU=$superusers{$_} NO=$nousers{$_} CL=$usecolour{$_} PR=$usepretty{$_}\n";
    }

  close userfile;
}

sub load_users
{
  my $userfile;

  logit "Server: Loading users";

  open(userfile, "<$usersfile");

  undef %aliases;
  undef %superusers;
  undef %userids;

  while (<userfile>)
    {
      if (/UN=(.*) AL=(.*) SU=(.*) NO=(.*) CL=(.*) PR=(.*)/)
	{
	  $aliases{$1} = $2;
	  $userids{lc $2} = $1;
	  $superusers{$1} = $3;
	  $nousers{$1} = $4;
	  $usecolour{$1} = $5;
	  $usepretty{$1} = $6;
	}
    }

  close userfile;
}

sub interrupt
  {
    for (values(%connection))
      {
	my($socket) = @$_[0];
	print $socket "\n\nSmelly server is going down NOW!!!! (Interrupt)\n";
	close $socket;
      }

    close $listener;
    die "Interrupt";
  }

sub pipe
{
  logit "Server: SIGPIPE";
}

$time_up = time();

load_users();
read_banned();
read_tracks();

set_volume();

my $weare = hostname;
$ourip = inet_ntoa(inet_aton($weare));
logit "Server: We are $weare ($ourip)";

if ($ourip =~ /^([0-9]+)\.([0-9]+)\.([0-9]+)\.([0-9]+)$/)
  {
    $ourip = [ $1, $2, $3, $4 ];
  }
else
  {
    logit "Server: Can't understand IP address";
    die;
  }

logit "Server: Starting (creating listening socket)...";
$listener = new IO::Socket::INET(LocalAddr => inet_ntoa(INADDR_ANY),
				 LocalPort => $listenport,
				 Proto     => 'tcp',
				 Listen    => 5,
				 Reuse     => 1,
				 Timeout   => 5);

die "Server: Can't create listening socket ($!)" unless $listener;

logit "Server: Starting (reading help information from data section)";

my $section = "";

while (<DATA>)
{
  if (/-- Welcome message/i)
    {
      $section = "welcome";
    }
  elsif (/-- Superuser help message/i)
    {
      $section = "suidhelp";
    }
  elsif (/-- Normal user help message/i)
    {
      $section = "normalhelp";
    }
  elsif (/^\*\*\*(.*)\s+$/)
    {
      if ($section eq "welcome")
	{
	  $intromsg[$#intromsg + 1] = "***$1";
	}
      elsif ($section eq "suidhelp")
	{
	  $suidhelp[$#suidhelp + 1] = "***$1";
	}
      elsif ($section eq "normalhelp")
	{
	  $helpmsg[$#helpmsg + 1] = "***$1";
	}
    }
}

sub sigchild
  {
    my $pid = waitpid(-1, &WNOHANG);
  }

my $no = $listener->fileno();
$sockets{$no} = [$listener, undef, \&pip];

logit "Server: Here we go..."; 

$SIG{INT} = \&interrupt;
$SIG{PIPE} = 'IGNORE';
$SIG{CHLD} = \&sigchild;

for (;;)
  {
    ping;
  }

__DATA__

# Assorted things & stuff

-- Welcome message
****************************************************************************
*** Welcome to the Slightly Smelly Server ('P-UKE')
*** Don't be naughty, or I'll cry (or crash)
*** Try .help for information
****************************************************************************
*** This version of the music server was written by Andrew Hunter
*** <esvce@dcs.warwick.ac.uk>. It has probably still got a few bugs in it,
*** so be nice to it.
****************************************************************************
*** Hello, %name%
****************************************************************************
-- Superuser help message
*** Note that you can identify users who are renaming themselves too much
*** by putting their real username in '%' signs - eg .kickban %csubj% hehe.
*** .urename has been fixed, too, although note that you use the 'real'
*** username for that - be careful not to rename the 'wrong' user when
*** someone has more than one username (the 'first' username is never
*** used)
***
*** Incidently, why do we always seem to have more superusers than ordinary
*** users?
****************************************************************************
-- Normal user help message
****************************************************************************
*** If you want to use random track queuing, speak to the resident bot,
*** Kraken (>kraken random on, for instance). .play, .list, etc use perl
*** regular expressions, not the old-style glob expressions. This creates
*** quite a lot more flexibility in picking tracks, but can be a bit more
*** confusing.
***
*** .play Foo_Fighters       - play everything by the foo_fighters (this is
***                            Kraken's hobby. He seems to like them)
*** .play (Radiohead|Indigo) - play everything with 'Radiohead' or 'Indigo'
***                            in the name
*** .play Foo[0-9]+Bar       - play anything with Foo followed by some 
***                            numbers followed by Bar
***
*** And so on...
****************************************************************************
