# Larry's brain

use IO::Socket;

# Describes what larry should do on certain events and actions

# ---
# The following functions should be provided:
#
# sub larry_say (who, what) - who has said what
# sub larry_tell(who, what) - who has told us what
#
# sub larry_user_start()                   - user list is starting
# sub larry_user      (who, uid, is_super) - user entry in user list
#
# sub larry_track_start() - list of available tracks is starting
# sub larry_track(track)  - reports available track in track list
#
# sub larry_queue_start(track,whoby) - start of track queue, playing track
# sub larry_queue_empty()            - reports empty queue
# sub larry_queue(pos, track, whoby) - track in queue
#
# sub larry_volume(volume)           - reports volume level
#
# sub larry_playing(track)           - track has started to play
# sub larry_emptied()                - queue has become empty
# sub larry_finished()               - playing has finished
# sub larry_stopped()                - current track has been stopped
#
# sub larry_queued(track, whoby)     - who has queued track
# sub larry_deleted(track, whoby)    - who has deleted track
# 
# sub larry_connected(who, suid)     - who has connected
# sub larry_disconn(who)             - who has disconnected
# sub larry_rename(who, whoto)       - who has renamed to whoto
# ---

# =======

my $sock;
my $botname;
my @bannedwords = (
		   "fuck", "shit", "bollocks", "gnome", "kde", "ibm", "censor",
		   "evolution", "darwin","harrow","camden","bromley","lewisham",
		   "fairy", "cock", "paul fox", "gary"
		  );

my @dontplay = (
		# "^.*\.mod\$"
	       );

# =======

my %suid; # maps usernames => suid status
my %uids; # maps usernames => real userids

my $filter = ".";
my @tracklist; # list of available tracks
my @playlist;
my %played;    # maps track names => (not)  played status

my $random = 0; # is randomness switched on?

my $queuelen = 0;
my $playing  = "";
my $playedby = "";

# =======

sub larry_init
  {
    my ($newsock, $bn) = @_;

    $sock = $newsock;
    $botname = $bn;
  }

# ========
# Larry version 1 actions

sub larry_say
  {
    my ($who, $what) = @_;

    if (lc($who) eq lc($botname))
      {
	return;
      }

    my $line = lc($what);

    for (@bannedwords)
      {
	my $check = $_;

	if ($line =~ /$check/)
	  {
	    print $sock ".kick $who You aren't allowed to say $check\n";

	    logit "Larry: kicking $who for using a banned word";
	  }
      }
  }

# =====

sub larry_user_start
  {
    # Forget about all users we used to know about
    undef %suid;
    undef %uids;
  }

sub larry_user
  {
    my ($who, $uid, $super) = @_;

    if ($super =~ /\{.*\}/)
      {
	$suid{$who} = 1;
      }
    else
      {
	$suid{$who} = 0;
      }

    $uids{$who} = $uid;
  }

sub larry_connected
  {
    my ($who, $super) = @_;

    if ($super =~ /\(.*\)/)
      {
	print $sock ": pokes $who\n";
      }

    print $sock ".who\n";
  }

sub larry_disconn
  {
    my ($who) = @_;

    if ($suid{$who})
      {
	print $sock "'Bye (sniff)\n";
      }

    print $sock ".who\n";
  }

sub larry_rename
  {
    print $sock ".who\n";
  }

# =====

sub larry_track_start
  {
    undef %played;
    undef @tracklist;
    undef @playlist;
  }

sub larry_track
  {
    my ($track) = @_;
    my $ok      = 1;

    for (@dontplay)
      {
	my $check = "if (\$track =~ /$_/) { \$ok = 0; }";

	eval $check;
	
	if ($@)
	  {
	    logit "Error: $check - $@";
	  }

	if (!$ok)
	  {
	    logit "Larry: not bothering with $track";
	    return;
	  }
      }

    push @tracklist, $track;
    my $x = 0;
    $x = eval "return \$track =~ /$filter/;";
    push @playlist, $track if $x == 1;
    $played{$track} = 1 if $x == 1;
#    logit "$track: $x\n";
  }

sub larry_reinit
  {
    print $sock ".list\n";
    logit "Larry: updating track list";
  }

# =====

sub larry_queue_start
  {
    my ($track, $whoby) = @_;

    $playing  = $track;
    $playedby = $whoby;
    $queuelen = 0;

    logit "Larry: played by $playedby";
  }

sub larry_queue_empty
  {
    $playing  = "";
    $queuelen = 0;

    if ($random == 1)
      {
	larry_playrandom();
      }
  }

sub larry_banned
  {
    larry_playrandom();
  }


sub larry_queue
  {
    my ($pos, $track, $whoby) = @_;

    if ($pos <= $queuelen || $pos > ($queuelen + 1))
      {
	logit "Larry: Hmm. Currently queuelen is $queuelen, but I get position $pos for next entry";
      }
    else
      {
	$queuelen = $pos;
	logit "Queue length $queuelen";
      }
  }

sub larry_queued
  {
    my ($track, $whoby) = @_;

    print $sock ".queue\n";
  }

sub larry_deleted
  {
    my ($track, $whoby) = @_;

    print $sock ".queue\n";
  }

sub larry_playing
  {
    my ($track) = @_;

    $played{$track} = 0;

    print $sock ".queue\n";
    print $sock "\n";
  }

sub larry_emptied
  {
    if ($random)
      {
	larry_playrandom();
      }
  }

sub larry_finished
  {
    print $sock ".queue\n";
    print $sock "\n";
  }

sub larry_stopped
  {
  }

# ====

sub larry_volume
  {
  }

# =====

# The commands larry understands
sub larry_tell
  {
    my ($who, $what) = @_;

    if ($what =~ /^\s*random\s+(on|off|now)\s*$/i)
      {
	my $state = lc($1);

	if ($state eq "on")
	  {
	    print $sock "'$who has switched on random track queuing\n";
	    $random = 1;

	    if ($queuelen < 1)
	      {
		larry_playrandom();
	      }
	  }
	elsif ($state eq "off")
	  {
	    print $sock "'$who has switched off random track queuing\n";
	    $random = 0;
	  }
	elsif ($state eq "now")
	  {
	    print $sock "'$who has requested a random track\n";
	    flush $sock;
	    larry_playrandom();
	  }
	else
	  {
	    print $sock "'$who managed to confuse me ($state)\n";
	  }
      }
    elsif ($what =~ /^\s*filter\s*([^\s]*)\s*$/) {
      $filter = $1;
      $filter =~ s/\//\\\//g;
      @foolist = map ($_, @tracklist);
      warn "Filtering /$filter/\n";
      print $sock ".say $who told me only to play from the list $filter\n";
      flush $sock;
      @playlist = eval "grep { /$filter/ } \@foolist; ";
      undef %played;
      foreach (@playlist) {
         $played{$_} = 1;
      }
      print "@playlist\n";
    }
    elsif ($what =~ /^\s*eat\s*(\w.*?)\s*$/) {
	my $food = $1;
	print $sock ".emote digests $food\n";
	if ($food =~ /engineering/) {
	    print $sock ".emote dies from food poisoning\n";
	    print $sock ".quit\n";
	} elsif ($food =~ /council/) {
	    print $sock ".emote projectile vomits over $who\n";
	} elsif ($food =~ /rms|esr|linus/i) {
	    print $sock ".emote becomes more powerful than you can possibly imagine\n";
	} elsif ($food =~ /harrow/i) {
	    print $sock ".emote strikes $who down for attempted poisoning\n";
	}
	flush $sock;
    }
    elsif ($what =~ /^\s*stop\s*$/)
      {
	if ($playedby eq $botname)
	  {
	    print $sock ".stop\n";
	    print $sock "'$who asked me to stop the current track\n";
	  }
	else
	  {
	    print $sock ">$who I didn't queue this track\n";
	  }
      }
    elsif ($what =~ /^\s*help\s*$/)
      {
	print $sock ">$who You can:\n";
	print $sock ">$who .tell Kraken random (on|off|now) - turns on or off random track queuing or queues a track now\n";
	print $sock ">$who .tell Kraken filter [regex] - only play tracks matching the regex\n";
	print $sock ">$who .tell Kraken eat [thing|person] - try it an see\n";
	  
	print $sock "\n";
	flush $sock;
	print $sock "\n";
      }
    else
      {
	if ($suid{$who})
	  {
	    if ($what =~ /^\s*do\s+(.*)\s*$/)
	      {
		print $sock ">$who OK, doing $1\n";
		print $sock "$1\n";
		print $sock "\n";
		flush $sock;
		print $sock "\n";
	      }
	    else
	      {
		print $sock ">$who I don't understand\n";
	      }
	  }
	else
	  {
	    print $sock ">$who I don't understand\n";
	  }
      }
  }

# =====

# Useful functions

sub larry_playrandom
  {
    # Check if all available tracks have been played
    my $total = 0;

    logit "Larry: Queuing random track";

    for (values(%played))
      {
	$total += $_;
      }

    if ($total == 0)
      {
	for (keys(%played))
	  {
	    $played{$_} = 1;
	  }
      }

    my ($num, $track);
    print "@playlist\n";
    do
      {
	$num = int(rand()*($#playlist + 1));
	$track = $playlist[$num];
      }
    while ($played{$track} == 0);

    logit "Larry: randomly queueing $track";
    print $sock ".play $track\n";

    print $sock "'$total tracks left\n";
    flush $sock;
    sleep 1;
    print $sock "\n";
  }

# =====

# Hmm
return 1;

